/* anim.c - Sprite animation module
 * 
 * Copyright 2001 by Robert J Ohannessian.
 *
 * This code is placed under the GNU General Public License.
 * Please refer to the accompanying file 'copying.txt' for details.
 */

#include <stdio.h>
#include <string.h>

#include "allegro.h"

#include "particle.h"
#include "game.h"

#include "log.h"


typedef struct ANIM_BANK {

	DATAFILE *datafile;

	BITMAP **bmp;
	int num_bitmaps;

	fixed cur_frame;
	fixed start_frame;
	fixed end_frame;
	fixed anim_speed;

	fixed angle;
	fixed rot_speed;

	int alpha;

	int prop;
	int life;
	int type;

	int osc_v_mag, osc_h_mag;
	fixed osc_v_phase, osc_v_speed;
	fixed osc_h_phase, osc_h_speed;
} ANIM_BANK;



static ANIM_BANK **AnimBank = NULL;
static char **AnimKeys = NULL;
static int NumAnims = 0;


static void interpret_anim(ANIM_BANK *anim, char *data) {

	char buf[512];
	char *end = data + ustrsizez(data) - 1;
	int i;
	float fl;

	if (!data || ustrsizez(data) < 2)
		return;

	while (sscanf(data, "%[^\n]", buf) > 0 && (data < end)) {
		if (buf[0] == '#') {
			data += ustrsizez(buf);
			continue;
		}

		if (sscanf(buf, "StartFrame=%i", &i) > 0) {
			anim->start_frame = itofix(i);
		}

		else if (sscanf(buf, "EndFrame=%i", &i) > 0) {
			anim->end_frame = itofix(i);
		}

		else if (sscanf(buf, "AnimSpeed=%f", &fl) > 0) {
			anim->anim_speed = ftofix(fl);
		}

		else if (sscanf(buf, "Loop=%i", &i) > 0) {
			if (i)
				anim->prop |= PARTICLE_LOOP_ANIM;
			else
				anim->prop &= ~PARTICLE_LOOP_ANIM;
		}

		else if (sscanf(buf, "DrawMethod=%s", buf + 256) > 0) {
			if (ustricmp("add", buf + 256) == 0)
				anim->prop |= PARTICLE_ALPHA;
			else
				anim->prop &= ~PARTICLE_ALPHA;
		}

		else if (sscanf(buf, "Oscillate=%s", buf + 256) > 0) {
			if (ustricmp("vert", buf + 256) == 0)
				anim->prop |= PARTICLE_OSCILLATE_V;
			else
				anim->prop &= ~PARTICLE_OSCILLATE_V;

			if (ustricmp("horiz", buf + 256) == 0)
				anim->prop |= PARTICLE_OSCILLATE_H;
			else
				anim->prop &= ~PARTICLE_OSCILLATE_H;

			if (ustricmp("both", buf + 256) == 0)
				anim->prop |= (PARTICLE_OSCILLATE_V | PARTICLE_OSCILLATE_H);
			else
				anim->prop &= ~(PARTICLE_OSCILLATE_V | PARTICLE_OSCILLATE_H);
		}

		else if (sscanf(buf, "OscillateVMag=%i", &i) > 0) {
			anim->osc_v_mag = i;
		}
		
		else if (sscanf(buf, "OscillateHMag=%i", &i) > 0) {
			anim->osc_h_mag = i;
		}

		else if (sscanf(buf, "OscillateVSpeed=%f", &fl) > 0) {
			anim->osc_v_speed = ftofix(fl);
		}
		
		else if (sscanf(buf, "OscillateHSpeed=%f", &fl) > 0) {
			anim->osc_h_speed = ftofix(fl);
		}

		else if (sscanf(buf, "OscillateVPhase=%f", &fl) > 0) {
			anim->osc_v_phase = ftofix(fl);
		}
		
		else if (sscanf(buf, "OscillateHPhase=%f", &fl) > 0) {
			anim->osc_h_phase = ftofix(fl);
		}

		data += ustrsizez(buf);
	}

	return;
}


static void add_bitmap(ANIM_BANK *anim, BITMAP *bmp) {

	BITMAP **t;

	if (!anim)
		return;

	if (anim->num_bitmaps) {
		if (bmp->w != anim->bmp[0]->w || bmp->h != anim->bmp[0]->h) {
			LOG("Error: Bitmaps not of same size.\n");
			return;
		}
	}
	
	t = realloc(anim->bmp, sizeof(BITMAP*) * (anim->num_bitmaps + 1));

	if (!t) {
		LOG("Error: Out of memory...\n");
		return;
	}
	anim->bmp = t;

	anim->bmp[anim->num_bitmaps] = bmp;
	anim->num_bitmaps++;

	return;
}

static ANIM_BANK *create_anim(DATAFILE *file) {

	ANIM_BANK *ret;

	ret = malloc(sizeof(ANIM_BANK));

	if (!ret)
		return NULL;

	memset(ret, 0, sizeof(ANIM_BANK));

	ret->datafile = file;

	return ret;
}


void add_anim(ANIM_BANK *anim, char *key) {

	ANIM_BANK **ab;
	char **kb;
	char *b;

	if (!anim || !key)
		return;

	ab = realloc(AnimBank, sizeof(ANIM_BANK*) * (NumAnims + 1));

	if (!ab)
		return;

	AnimBank = ab;

	kb = realloc(AnimKeys, sizeof(char*) * (NumAnims + 1));

	if (!kb)
		return;

	AnimKeys = kb;


	b = malloc(sizeof(char) * ustrsizez(key));

	if (!b)
		return;

	ustrcpy(b, key);

	kb[NumAnims] = b;
	ab[NumAnims] = anim;

	NumAnims++;
}


void load_anim_bank(char *key, char *filename) {

	DATAFILE *data;
	int i;
	ANIM_BANK *anim;

	if (!key || !filename)
		return;

	data = load_datafile(filename);

	if (!data) {
		LOG("Unable to open %s\n", filename);
		return;
	}

	anim = create_anim(data);

	/* Look for ANI file */
	for (i = 0; ; i++) {
		
		if (data[i].type == DAT_END)
			break;

		if (data[i].type == AL_ID('A','N','I','M'))
			interpret_anim(anim, data[i].dat);

		if (data[i].type == DAT_BITMAP) {
			add_bitmap(anim, data[i].dat);
		}
	}

	add_anim(anim, key);

	return;
}


PARTICLE *make_anim_particle(char *key) {

	int i;
	PARTICLE *p;
	ANIM_BANK *anim = NULL;

	if (!key)
		return NULL;

	/* Find key */
	for (i = 0; i < NumAnims; i++) {
		if (AnimKeys[i] && stricmp(key, AnimKeys[i]) == 0) {
			anim = AnimBank[i];
			break;
		}
	}

	if (i >= NumAnims || !anim)
		return NULL;

	p = create_particle();

	for (i = 0; i < anim->num_bitmaps; i++)
		add_particle_bitmap(p, anim->bmp[i]);

	set_particle_frame(p, fixtof(anim->cur_frame));
	set_particle_frames(p, fixtof(anim->start_frame), fixtof(anim->end_frame));
	set_particle_anim_speed(p, fixtof(anim->anim_speed));
	set_particle_angle(p, fixtof(anim->angle));
	set_particle_rotation_speed(p, anim->rot_speed);
	set_particle_alpha(p, anim->alpha);
	set_particle_property(p, anim->prop);
	set_particle_life(p, anim->life);
	set_particle_type(p, anim->type);

	set_particle_osc_v(p, anim->osc_v_mag, fixtof(anim->osc_v_speed));
	set_particle_osc_h(p, anim->osc_h_mag, fixtof(anim->osc_h_speed));		

	return p;
}

void remove_anim() {
	int i;

	for (i = 0; i < NumAnims; i++) {

		if (AnimKeys[i])
			free(AnimKeys[i]);

		if (AnimBank[i]->datafile)
			unload_datafile(AnimBank[i]->datafile);
		if (AnimBank[i]->bmp)
			free(AnimBank[i]->bmp);

		free(AnimBank[i]);
	}

	if (AnimBank)
		free(AnimBank);
	if (AnimKeys)
		free(AnimKeys);

	NumAnims = 0;
	AnimBank = NULL;
	AnimKeys = NULL;

	return;
}
