/* gametimer.c - Game timing module
 * 
 * Copyright 2001 by Robert J Ohannessian.
 *
 * This code is placed under the GNU General Public License.
 * Please refer to the accompanying file 'copying.txt' for details.
 */
#include "allegro.h"
#include "gfx.h"
#include "gametimer.h"
#include "game.h"
#include "log.h"


char *ErrorString[GTIME_MAX_ERRORS] = {
	"No Errors",
	"Out of memory while allocating %i bytes",
	"Null argument '%s'",
	"GameTimer Not installed"
};


/* Notes: There is NO guarenty on the order of the calls issued by Poll */
/* Plus, I should speed it up by allocating blocks of the array at a time */

volatile int GameTime = 0;
static int GameTimerInstalled = FALSE;


typedef struct GTIME_POLL {
	int (*Function)(void*);
	int Period;
	int Mode;
	int Timer;
	void *Data;
} GTIME_POLL;

static GTIME_POLL *Polled = NULL;
static int NumPolled = 0;
static int GTIME_FreezeFlag = FALSE;

void GTIME_Handler(void) {

	if (!GTIME_FreezeFlag)
		GameTime++;

} END_OF_FUNCTION(GTIME_Handler);

void GTIME_Remove(void) {

	if (!GameTimerInstalled)
		VERROR(GTIME_ERROR_NOT_INSTALLED, 0);

	remove_int(GTIME_Handler);

	if (Polled)
		free(Polled);
	Polled = NULL;
	NumPolled = 0;

	GameTimerInstalled = FALSE;
}

int GTIME_Install() {

	if (GameTimerInstalled)
		ERROR(GTIME_ERROR_ALREADY_INSTALLED, 0);

	LOCK_VARIABLE(GameTime);
	LOCK_FUNCTION(GTIME_Handler);

	install_int_ex(GTIME_Handler, BPS_TO_TIMER(GAMETIMER_FREQUENCY));
	atexit(GTIME_Remove);

	GameTimerInstalled = TRUE;

	return TRUE;
}

int GTIME_AddFunction(int (*func)(void*), int period, void *data, int mode) {

	GTIME_POLL *t;

	if (!func)
		ERROR(GTIME_ERROR_NULL_ARGUMENT, "func");
	if (!period)
		ERROR(GTIME_ERROR_NULL_ARGUMENT, "period");

	t = realloc(Polled, sizeof(GTIME_POLL) * (NumPolled + 1));
	
	if (!t)
		ERROR(GTIME_ERROR_OUT_OF_MEMORY, sizeof(GTIME_POLL) * (NumPolled + 1));

	Polled = t;
	Polled[NumPolled].Function = func;
	Polled[NumPolled].Data = data;
	Polled[NumPolled].Period = period;
	Polled[NumPolled].Timer = GameTime;
	Polled[NumPolled].Mode = mode;

	NumPolled++;

	return GTIME_OK;
}
	
int GTIME_RemoveFunction(int (*func)(void*), void *data) {

	int i;
	GTIME_POLL *t;

	if (!func)
		ERROR(GTIME_ERROR_NULL_ARGUMENT, "func");

	for (i = 0; i < NumPolled; i++) {

		if (Polled[i].Function == func && Polled[i].Data == data) {

			Polled[i] = Polled[NumPolled - 1];
			NumPolled--;

			if (!NumPolled) {
				free(Polled);
				Polled = NULL;
				break;
			}
			else {
				t = realloc(Polled, sizeof(GTIME_POLL) * NumPolled);

				if (!t)
					ERROR(GTIME_ERROR_OUT_OF_MEMORY, sizeof(GTIME_POLL) * NumPolled);

				Polled = t;
			}
			i--;
			continue;
		}
	}
	return GTIME_OK;
}

int GTIME_Reset() {

	int i;

	for (i = 0; i < NumPolled; i++)
		Polled[i].Timer = GameTime;

	return GTIME_OK;
}

int GTIME_Poll() {

	int the_time, i, r, timers_elapsed = 0;

	/* We have todo this in a function which gets called from both
	   the game loop and the menu loop hence we do it here. Not a logical
	   place or pretty, but it works. */
	if (FullscreenToggleRequested)
	{
		fullscreen = !fullscreen;
		if (set_gfx_mode(fullscreen? GFX_AUTODETECT_FULLSCREEN:
				GFX_AUTODETECT_WINDOWED,
				DISPLAY_W, DISPLAY_H, 0, 0))
		{
			fprintf(stderr,
				"FATAL error switching Fullscreen<->Window\n");
			exit(1);
		}
		set_palette(GamePal);
		FullscreenToggleRequested = 0;
	}
	
	the_time = GameTime;
	
	for (i = 0; i < NumPolled; i++) {

		GTIME_POLL *p = Polled + i;		
		int time_delta = the_time - p->Timer;
		r = 0;

		if (time_delta > p->Period) {

			if (p->Mode & GTIME_FAST) {
				p->Timer += (time_delta / p->Period) * p->Period;
				if (p->Function) {
					int old_time = GameTime;
					GameTime = p->Timer;
					r = p->Function(p->Data);
					GameTime = old_time;
				}
			}
			else {
				time_delta /= p->Period;
				while (time_delta) {
					p->Timer += p->Period;
					if (p->Function) {
						r = p->Function(p->Data);
					}
					if (r)
						break;
					time_delta--;					
				}
			}

			if (r) {
				i--;
				continue;
			}
			timers_elapsed++;
		}
	}

	if (timers_elapsed == 0)
		rest(1);
		
	return GTIME_OK;
}

int GTIME_FreezeTimer() {

	GTIME_FreezeFlag = TRUE;
	return GTIME_OK;
}

int GTIME_UnFreezeTimer() {

	GTIME_FreezeFlag = FALSE;
	return GTIME_OK;
}

int GTIME_IsFrozen() {
	return GTIME_FreezeFlag;
}
