/* init.c - Player input management module
 * 
 * Copyright 2001 by Robert J Ohannessian.
 *
 * This code is placed under the GNU General Public License.
 * Please refer to the accompanying file 'copying.txt' for details.
 */

#include <string.h>

#include "allegro.h"
#include "input.h"
#include "log.h"


static const char *scancode_name[] = 
{
   "(none)",     "A",          "B",          "C",
   "D",          "E",          "F",          "G",
   "H",          "I",          "J",          "K",
   "L",          "M",          "N",          "O",
   "P",          "Q",          "R",          "S",
   "T",          "U",          "V",          "W",
   "X",          "Y",          "Z",          "0",
   "1",          "2",          "3",          "4",
   "5",          "6",          "7",          "8",
   "9",          "0_PAD",      "1_PAD",      "2_PAD",
   "3_PAD",      "4_PAD",      "5_PAD",      "6_PAD",
   "7_PAD",      "8_PAD",      "9_PAD",      "F1",
   "F2",         "F3",         "F4",         "F5",
   "F6",         "F7",         "F8",         "F9",
   "F10",        "F11",        "F12",        "ESC",
   "TILDE",      "MINUS",      "EQUALS",     "BACKSPACE",
   "TAB",        "OPENBRACE",  "CLOSEBRACE", "ENTER",
   "COLON",      "QUOTE",      "BACKSLASH",  "BACKSLASH2",
   "COMMA",      "STOP",       "SLASH",      "SPACE",
   "INSERT",     "DEL",        "HOME",       "END",
   "PGUP",       "PGDN",       "LEFT",       "RIGHT",
   "UP",         "DOWN",       "SLASH_PAD",  "ASTERISK",
   "MINUS_PAD",  "PLUS_PAD",   "DEL_PAD",    "ENTER_PAD",
   "PRTSCR",     "PAUSE",      "ABNT_C1",    "YEN",
   "KANA",       "CONVERT",    "NOCONVERT",  "AT",
   "CIRCUMFLEX", "COLON2",     "KANJI",	     
   "EQUALS_PAD", "BACKQUOTE",  "SEMICOLON",  "COMMAND",
   "UNKNOWN1",   "UNKNOWN2",   "UNKNOWN3",   "UNKNOWN4",
   "UNKNOWN5",   "UNKNOWN6",   "UNKNOWN7",   "UNKNOWN8",
   "LSHIFT",     "RSHIFT",     "LCONTROL",   "RCONTROL",
   "ALT",        "ALTGR",      "LWIN",       "RWIN",
   "MENU",       "SCRLOCK",    "NUMLOCK",    "CAPSLOCK",
   "MAX"
};

static char *input_name[] = {
	"(none)", "Keyboard", "Mouse", "Joystick"
};


INPUT *create_input(int dev) {

	INPUT *in;

	in = malloc(sizeof(INPUT));

	if (!in) {
		LOG("--== Error ==-- Input::create_input: Out of memory while allocating %i bytes\n", sizeof(INPUT));
		return NULL;
	}

	if (dev == INPUT_TYPE_KEYBOARD) {
		in->InputData = malloc(sizeof(INPUT_KEYB));
		in->InputType = INPUT_TYPE_KEYBOARD;

		if (!in->InputData) {
			free(in);
			LOG("--== Error ==-- Input::create_input: Out of memory while allocating %i bytes\n", sizeof(INPUT_KEYB));
			return NULL;
		}

		memset(in->InputData, 0, sizeof(INPUT_KEYB));
	}
	else if (dev == INPUT_TYPE_MOUSE) {
		in->InputData = malloc(sizeof(INPUT_MOUSE));
		in->InputType = INPUT_TYPE_MOUSE;

		if (!in->InputData) {
			free(in);
			LOG("--== Error ==-- Input::create_input: Out of memory while allocating %i bytes\n", sizeof(INPUT_MOUSE));
			return NULL;
		}

		memset(in->InputData, 0, sizeof(INPUT_MOUSE));
	}
	else if (dev == INPUT_TYPE_JOYSTICK) {
		in->InputData = malloc(sizeof(INPUT_JOY));
		in->InputType = INPUT_TYPE_JOYSTICK;

		if (!in->InputData) {
			free(in);
			LOG("--== Error ==-- Input::create_input: Out of memory while allocating %i bytes\n", sizeof(INPUT_JOY));
			return NULL;
		}

		memset(in->InputData, 0, sizeof(INPUT_JOY));
	}
	else {
		LOG("--== Error ==-- Input::create_input: 'dev' is not of any known value (%i)\n", dev);
		free(in);
		return NULL;
	}


	return in;
}


void destroy_input(INPUT **in) {

	if (!in) {
		LOG("--== Error ==-- Input::destroy_input: Null argument 'in'\n");
		return;
	}
	if (!*in) {
		LOG("--== Error ==-- Input::destroy_input: Null argument '*in'\n");
		return;
	}

	if ((*in)->InputData)
		free((*in)->InputData);

	free(*in);
	*in = NULL;

	return;
}


static int get_joystick_axis(int jn, int sn, int an) {

	if (an < 0)
		return joy[jn].stick[sn].axis[-an - 1].d1;
	else
		return joy[jn].stick[sn].axis[an].d2;
}

static int get_joystick_button(int jn, int bn) {

		return joy[jn].button[bn].b;
}


int get_input(INPUT *dev, int k) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input: Null argument 'dev'\n");
		return 0;
	}
	if (!k) {
		LOG("--== Error ==-- Input::get_input: Null argument 'k'\n");
		return 0;
	}

	if (dev->InputType == INPUT_TYPE_KEYBOARD) {

		INPUT_KEYB *keyb = (INPUT_KEYB*)dev->InputData;

		if (!keyb) {
			LOG("--== Error ==-- Input::get_input: dev->InputData is NULL\n");
			return 0;
		}

		return ((k & INPUT_KEY_UP)    && key[keyb->Up])
			|| ((k & INPUT_KEY_DOWN)  && key[keyb->Down])
			|| ((k & INPUT_KEY_LEFT)  && key[keyb->Left])
			|| ((k & INPUT_KEY_RIGHT) && key[keyb->Right])
			|| ((k & INPUT_KEY_FIRE)  && key[keyb->Fire])
			;
	}
	else if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		INPUT_JOY *joys = (INPUT_JOY*)dev->InputData;
		int ret = FALSE;

		if (!joys) {
			LOG("--== Error ==-- Input::get_input: dev->InputData is NULL\n");
			return 0;
		}

		/* Axis */
		if (k & INPUT_KEY_UP)
			ret |= get_joystick_axis(joys->JoystickNum, joys->UpStick, joys->UpAxis);
		if (k & INPUT_KEY_DOWN)
			ret |= get_joystick_axis(joys->JoystickNum, joys->DownStick, joys->DownAxis);
		if (k & INPUT_KEY_RIGHT)
			ret |= get_joystick_axis(joys->JoystickNum, joys->RightStick, joys->RightAxis);
		if (k & INPUT_KEY_LEFT)
			ret |= get_joystick_axis(joys->JoystickNum, joys->LeftStick, joys->LeftAxis);
		if (k & INPUT_KEY_FIRE)
			ret |= get_joystick_button(joys->JoystickNum, joys->Fire);

		return ret;
	}
	return 0;
}


void set_input_key(INPUT *dev, int control, int k) {


	if (!dev) {
		LOG("--== Error ==-- Input::set_input_key: Null argument 'dev'\n");
		return;
	}

	if (dev->InputType == INPUT_TYPE_KEYBOARD) {

		INPUT_KEYB *keyb = (INPUT_KEYB*)dev->InputData;

		if (!keyb) {
			LOG("--== Error ==-- Input::set_input_key: dev->InputData is NULL\n");
			return;
		}
	
		if (control & INPUT_KEY_UP)
			keyb->Up = k;
		if (control & INPUT_KEY_DOWN)
			keyb->Down = k;
		if (control & INPUT_KEY_RIGHT)
			keyb->Right = k;
		if (control & INPUT_KEY_LEFT)
			keyb->Left = k;
		if (control & INPUT_KEY_FIRE)
			keyb->Fire = k;		
	}
	else
		LOG("--== Error ==-- Input::set_input_key: Attempting to set keyboard keys on a non-keyboard device (%i).\n", dev->InputType);

	return;
}


void set_input_joy(INPUT *dev, int control, int jn, int sn, int an) {

	if (!dev) {
		LOG("--== Error ==-- Input::set_input_joy: Null argument 'dev'\n");
		return;
	}

	if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		INPUT_JOY *joys = (INPUT_JOY*)dev->InputData;

		if (!joys) {
			LOG("--== Error ==-- Input::set_input_joy: dev->InputData is NULL\n");
			return;
		}

		if (jn >= 0)
			joys->JoystickNum = jn;

		if (control & INPUT_KEY_UP) {
			joys->UpAxis = an;
			joys->UpStick = sn;
		}
		else if (control & INPUT_KEY_DOWN) {
			joys->DownAxis = an;
			joys->DownStick = sn;
		}
		else if (control & INPUT_KEY_RIGHT) {
			joys->RightAxis = an;
			joys->RightStick = sn;
		}
		else if (control & INPUT_KEY_LEFT) {
			joys->LeftAxis = an;
			joys->LeftStick = sn;
		}

		else if (control & INPUT_KEY_FIRE) {
			joys->Fire = sn;
		}		
	}
	else
		LOG("--== Error ==-- Input::set_input_joy: Attempting to set joystick keys on a non-joystick device (%i).\n", dev->InputType);

	return;
}



int get_input_type(INPUT *dev) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_type: Null argument 'dev'\n");
		return 0;
	}

	return dev->InputType;
}

char* get_input_type_name(INPUT *dev) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_type: Null argument 'dev'\n");
		return 0;
	}

	if (dev->InputType < 0 || dev->InputType >= INPUT_TYPE_MAX) {
		LOG("--== Error ==-- Input::get_input_type: Incorrect value for input type (%i).\n", dev->InputType);
		return 0;
	}

	return input_name[dev->InputType];
}


char* get_input_type_full_name(INPUT *dev) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_type: Null argument 'dev'\n");
		return 0;
	}

	if (dev->InputType < 0 || dev->InputType >= INPUT_TYPE_MAX) {
		LOG("--== Error ==-- Input::get_input_type: Incorrect value for input type (%i).\n", dev->InputType);
		return 0;
	}

	if (dev->InputType == INPUT_TYPE_JOYSTICK) {
		INPUT_JOY *joys = dev->InputData;

		if (joys) {
			static char buf[16];
			snprintf(buf, sizeof(buf), "Joystick %i",
				joys->JoystickNum + 1);
			return buf;
		}
	}

	return input_name[dev->InputType];
}


int get_input_key(INPUT *dev, int control) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_key: Null argument 'dev'\n");
		return 0;
	}

	if (dev->InputType == INPUT_TYPE_KEYBOARD) {

		INPUT_KEYB *keyb = (INPUT_KEYB*)dev->InputData;

		if (!keyb) {
			LOG("--== Error ==-- Input::get_input_key: dev->InputData is NULL\n");
			return 0;
		}

		if (control & INPUT_KEY_UP)
			return keyb->Up;
		if (control & INPUT_KEY_DOWN)
			return keyb->Down;
		if (control & INPUT_KEY_RIGHT)
			return keyb->Right;
		if (control & INPUT_KEY_LEFT)
			return keyb->Left;
		if (control & INPUT_KEY_FIRE)
			return keyb->Fire;
	}
	else
		LOG("--== Error ==-- Input::get_input_key: Attempting to get keyboard data out of a non-keyboard device (%i).\n", dev->InputType);

	return 0;
}

int get_input_joy_button(INPUT *dev, int control) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_key: Null argument 'dev'\n");
		return 0;
	}

	if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		INPUT_JOY *joys = (INPUT_JOY*)dev->InputData;

		if (!joys) {
			LOG("--== Error ==-- Input::get_input_joy_button: dev->InputData is NULL\n");
			return 0;
		}

		if (control & INPUT_KEY_FIRE)
			return joys->Fire;
	}
	else
		LOG("--== Error ==-- Input::get_input_joy_button: Attempting to get joystick data out of a non-joystick device (%i).\n", dev->InputType);

	return 0;
}


int get_input_joy_stick(INPUT *dev, int control) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_joy_stick: Null argument 'dev'\n");
		return 0;
	}

	if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		INPUT_JOY *joys = (INPUT_JOY*)dev->InputData;

		if (!joys) {
			LOG("--== Error ==-- Input::get_input_joy_button: dev->InputData is NULL\n");
			return 0;
		}

		if (control & INPUT_KEY_UP)
			return joys->UpStick;
		if (control & INPUT_KEY_DOWN)
			return joys->DownStick;
		if (control & INPUT_KEY_RIGHT)
			return joys->RightStick;
		if (control & INPUT_KEY_LEFT)
			return joys->LeftStick;
	}
	else
		LOG("--== Error ==-- Input::get_input_joy_stick: Attempting to get joystick data out of a non-joystick device (%i).\n", dev->InputType);

	return 0;
}


int get_input_joy_axis(INPUT *dev, int control) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_joy_axis: Null argument 'dev'\n");
		return 0;
	}

	if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		INPUT_JOY *joys = (INPUT_JOY*)dev->InputData;

		if (!joys) {
			LOG("--== Error ==-- Input::get_input_joy_axis: dev->InputData is NULL\n");
			return 0;
		}

		if (control & INPUT_KEY_UP)
			return joys->UpAxis;
		if (control & INPUT_KEY_DOWN)
			return joys->DownAxis;
		if (control & INPUT_KEY_RIGHT)
			return joys->RightAxis;
		if (control & INPUT_KEY_LEFT)
			return joys->LeftAxis;
	}
	else
		LOG("--== Error ==-- Input::get_input_joy_axis: Attempting to get joystick data out of a non-joystick device (%i).\n", dev->InputType);

	return 0;
}


const char *get_input_name(INPUT *dev, int control, char *buf, int length) {

	if (!dev) {
		LOG("--== Error ==-- Input::get_input_name: Null argument 'dev'\n");
		return NULL;
	}
	if (!buf) {
		LOG("--== Error ==-- Input::get_input_name: Null argument 'buf'\n");
		return NULL;
	}


	if (length && length < 11) {
		LOG("--== Error ==-- Input::get_input_name: Buffer length (%i) must be at least 11 bytes long.\n");
		return NULL;
	}
		

	if (dev->InputType == INPUT_TYPE_KEYBOARD) {

		const char *r = NULL;
		INPUT_KEYB *keyb = (INPUT_KEYB*)dev->InputData;		

		if (!keyb) {
			LOG("--== Error ==-- Input::get_input_name: dev->InputData is NULL\n");
			return NULL;
		}

		if (control & INPUT_KEY_UP)
			r = scancode_name[keyb->Up];
		else if (control & INPUT_KEY_DOWN)
			r = scancode_name[keyb->Down];
		else if (control & INPUT_KEY_RIGHT)
			r = scancode_name[keyb->Right];
		else if (control & INPUT_KEY_LEFT)
			r = scancode_name[keyb->Left];
		else if (control & INPUT_KEY_FIRE)
			r = scancode_name[keyb->Fire];

		if (!r) {
			LOG("--== Error ==- Input::get_input_name: Invalid key binding for 'control' (%i).\n", control);
			return NULL;
		}

		if (length)
		{
			sprintf(buf, "%*s", length - 1, r);
			return buf;
		}
		else
			return r;
	}

	else if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		const char *r = NULL;
		int an = -1, sn = -1;
		INPUT_JOY *joys = (INPUT_JOY*)dev->InputData;
		int nse = 0;

		if (!joys) {
			LOG("--== Error ==-- Input::get_input_name: dev->InputData is NULL\n");
			return NULL;
		}

		if (control & INPUT_KEY_UP) {
			an = joys->UpAxis;
			sn = joys->UpStick;
			r = joy[joys->JoystickNum].stick[sn].axis[an < 0 ? -an - 1 : an].name;
		}
		else if (control & INPUT_KEY_DOWN) {
			an = joys->DownAxis;
			sn = joys->DownStick;
			r = joy[joys->JoystickNum].stick[sn].axis[an < 0 ? -an - 1 : an].name;
		}
		else if (control & INPUT_KEY_RIGHT) {
			an = joys->RightAxis;
			sn = joys->RightStick;
			r = joy[joys->JoystickNum].stick[sn].axis[an < 0 ? -an - 1 : an].name;
		}
		else if (control & INPUT_KEY_LEFT) {
			an = joys->LeftAxis;
			sn = joys->LeftStick;
			r = joy[joys->JoystickNum].stick[sn].axis[an < 0 ? -an - 1 : an].name;
		}

		else if (control & INPUT_KEY_FIRE) {
			an = 0;
			sn = joys->Fire;
			r = joy[joys->JoystickNum].button[sn].name;
			nse = 1;
		}

		if (control && sn == -1) {
			LOG("--== Error ==- Input::get_input_name: Invalid key binding for 'control' (%i).\n", control);
			return NULL;
		}

		/* Print minimal form? */
		if (length == 0)
			sprintf(buf, "Joy%i %i %i", joys->JoystickNum, sn, an);
		else
			sprintf(buf, "%*s%c", length - 1, r, nse ? 0 : ((an < 0) ? '+' : '-'));

		return buf;
	}
	else
		LOG("--== Error ==-- Input::get_input_name: Attempting to get name data out of an unknown device (%i).\n", dev->InputType);

	return NULL;
}


int set_input(INPUT *dev, int control, const char *k) {

	int i;

	if (!dev) {
		LOG("--== Error ==-- Input::set_input: Null argument 'dev'\n");
		return -1;
	}
	if (!k) {
		LOG("--== Error ==-- Input::set_input: Null argument 'k'\n");
		return -1;
	}

	if (dev->InputType == INPUT_TYPE_KEYBOARD) {

		int new_key = 0;

		for (i = 0; i < KEY_MAX; i++) {
			if (stricmp(k, scancode_name[i]) == 0) {
				new_key = i;
				break;
			}
		}

		if (!new_key) {
			LOG("--== Warning ==-- Input::set_input: New key binding is unaccepted (%s).\n", k);
			return 1;
		}

		set_input_key(dev, control, new_key);
	}
	else if (dev->InputType == INPUT_TYPE_JOYSTICK) {

		int n;
		int jn, an, sn;

		n = sscanf(k, "Joy%i %i %i", &jn, &sn, &an);

		if (n != 3) {
			LOG("--== Warning ==-- Input::set_input: New joystick binding is unaccepted (%s).\n", k);
			return 1;
		}

		/* Sanity check for user inputted values */
		if (jn >= num_joysticks) {
			LOG("--== Error ==-- Input::set_input: Attempting to bind to undefined joystick (%i).\n", jn);
			return -1;
		}
		if ((sn >= joy[jn].num_buttons) && (control & INPUT_KEY_FIRE)) {
			LOG("--== Error ==-- Input::set_input: Attempting to bind to undefined joystick button (%i).\n", sn);
			return -1;
		}
		if ((sn >= joy[jn].num_sticks) && ((control & INPUT_KEY_FIRE) == 0)) {
			LOG("--== Error ==-- Input::set_input: Attempting to bind to undefined joystick stick (%i).\n", sn);
			return -1;
		}
		if (((control & INPUT_KEY_FIRE) == 0) && ((an >= joy[jn].stick[sn].num_axis) || ((-an - 1) >= joy[jn].stick[sn].num_axis))) {
			LOG("--== Error ==-- Input::set_input: Attempting to bind to undefined joystick axis(%i).\n", an);
			return -1;
		}

		set_input_joy(dev, control, jn, sn, an);
	}
	else
		LOG("--== Error ==-- Input::set_input: Attempting to set binding data for an unknown device (%i).\n", dev->InputType);
	
	return 0;
}
