/* particle.c - Particle system manager module
 * 
 * Copyright 2001 by Robert J Ohannessian.
 *
 * This code is placed under the GNU General Public License.
 * Please refer to the accompanying file 'copying.txt' for details.
 */

#include <string.h>

#include "allegro.h"
#include "gametimer.h"
#include "particle.h"
#include "ppcol.h"
#include "gfx.h"
#include "log.h"
#include "game.h"
#include "drs.h"


#ifndef M_PI
	#define M_PI 3.14159
#endif


PARTICLE **Particle = NULL;
int NumParticles = 0;
static int ArraySize = 0;
static int ParticleInstalled = 0;


static void add_particle(PARTICLE *p) {

	if ((ArraySize - NumParticles) <= 1) {
		PARTICLE **part;

		part = realloc(Particle, sizeof(PARTICLE*) * (ArraySize + 32));

		if (!part) {
			LOG("--== Error ==-- Particle::add_particle: Ran out of memory while allocating %i bytes.\n", sizeof(PARTICLE*) * (ArraySize + 32));
			return;
		}

		Particle = part;
		ArraySize += 32;
	}

	Particle[NumParticles] = p;

	NumParticles++;

	return;
}


static void remove_particle(PARTICLE *p) {

	int i;


	for (i = 0; i < NumParticles; i++) {
		if (Particle[i] == p) {		
			Particle[i] = Particle[NumParticles - 1];
			NumParticles--;
			break;
		}
	}

	if (!NumParticles) {
		if (Particle)
			free(Particle);
		Particle = NULL;
		ArraySize = 0;
	}
	else if ((ArraySize - NumParticles) > 32) {
		ArraySize -= 32;
		ArraySize = MAX(32, ArraySize);
		Particle = realloc(Particle, sizeof(PARTICLE*) * ArraySize);

		if (!Particle)
			LOG("--== Error ==-- Particle::remove_particle: Unable to resize block to %i bytes.\n", sizeof(PARTICLE*) * ArraySize);			
	}
	return;
}

int find_particle(PARTICLE *p) {
	int i;

	for (i = 0; i < NumParticles; i++) {
		if (Particle[i] == p) {		
			return TRUE;
		}
	}

	return FALSE;
}

PARTICLE *create_particle() {

	PARTICLE *p = malloc(sizeof(PARTICLE));

	if (!p) {
		LOG("--== Error ==-- Particle::create_particle: Ran out of memory while allocating %i bytes.\n", sizeof(PARTICLE));
		return NULL;
	}

	memset(p, 0, sizeof(PARTICLE));
	p->alpha = 255;
	p->old_x = -9999999;

	add_particle(p);

	return p;
}

void set_particle_coord(PARTICLE *p, float x, float y) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_coord: Parameter 'p' is NULL.\n");
		return;
	}

	p->x = x;
	p->y = y;

	return;
}

void set_particle_velocity(PARTICLE *p, float vx, float vy) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_velocity: Parameter 'p' is NULL.\n");
		return;
	}

	p->vx = vx;
	p->vy = vy;

	return;
}

void set_particle_angle(PARTICLE *p, float angle) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_angle: Parameter 'p' is NULL.\n");
		return;
	}	

	p->angle = ftofix(angle / 3.14159 * 128);

	return;
}

void set_particle_rotation_speed(PARTICLE *p, fixed rot_speed) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_rotation_speed: Parameter 'p' is NULL.\n");
		return;
	}
	
	p->rot_speed = rot_speed;

	return;
}

void set_particle_alpha(PARTICLE *p, int a) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_alpha: Parameter 'p' is NULL.\n");
		return;
	}
	
	p->alpha = a;

	return;
}

void add_particle_bitmap(PARTICLE *p, BITMAP *bmp) {

	BITMAP **t;

	if (!p) {
		LOG("--== Error ==-- Particle::add_particle_bitmap: Parameter 'p' is NULL.\n");
		return;
	}

	if (p->num_bitmaps > 0) {
		if (bmp->w != p->bmp[0]->w || bmp->h != p->bmp[0]->h)
			return;
	}

	t = realloc(p->bmp, sizeof(BITMAP*) * (p->num_bitmaps + 1));

	if (!t) {
		LOG("--== Error ==-- Particle::add_particle_bitmap: Ran out of memory while allocating %i bytes\n", sizeof(BITMAP*) * (p->num_bitmaps + 1));
		return;
	}

	p->bmp = t;


	p->bmp[p->num_bitmaps] = bmp;
	p->num_bitmaps++;

	if (p->aux_bmp)
		destroy_bitmap(p->aux_bmp);

	p->aux_bmp = NULL;

	return;
}

void remove_particle_bitmaps(PARTICLE *p) {

	if (!p) {
		LOG("--== Error ==-- Particle::remove_particle_bitmap: Parameter 'p' is NULL.\n");
		return;
	}

	if (!p->bmp) {
		LOG("--== Error ==-- Particle::remove_particle_bitmap: Parameter 'p->bmp' is NULL.\n");
		return;
	}

	free(p->bmp);
	p->bmp = NULL;
	p->num_bitmaps = 0;
	p->cur_frame = 0;

	if (p->aux_bmp) {
		int x = p->old_x,
			y = p->old_y,
			w = p->aux_bmp->w,
			h = p->aux_bmp->h;

		DRS_add_rectangle(x, y, w, h);
		DRS_add_rectangle(x + DISPLAY_W, y, w, h);
		DRS_add_rectangle(x - DISPLAY_W, y, w, h);
		DRS_add_rectangle(x, y + DISPLAY_H, w, h);
		DRS_add_rectangle(x, y - DISPLAY_H, w, h);
	}

	if (p->aux_bmp)
		destroy_bitmap(p->aux_bmp);

	p->aux_bmp = NULL;

	return;
}

void set_particle_life(PARTICLE *p, int val) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_life: Parameter 'p' is NULL.\n");
		return;
	}

	p->life = val;

	return;
}

void set_particle_type(PARTICLE *p, int val) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_type: Parameter 'p' is NULL.\n");
		return;
	}

	p->type = val;

	return;
}


void set_particle_property(PARTICLE *p, int val) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_property: Parameter 'p' is NULL.\n");
		return;
	}
	
	p->prop = val;
}


float get_particle_angle(PARTICLE *p) {

	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_angle: Parameter 'p' is NULL.\n");
		return 0;
	}

	return fixtof(p->angle & 0xFFFFFF) * 3.14159 / 128;
}

void set_particle_owner(PARTICLE *p, PARTICLE *owner) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_owner: Parameter 'p' is NULL.\n");
		return;
	}

	p->owner = owner;

	return;
}

void destroy_particle(PARTICLE *p) {

	if (!find_particle(p)) {
		LOG("--== Error ==-- Particle::destroy_particle: Unable to find particle!\n");
		return;
	}

	if (p->aux_bmp) {
		int x = p->old_x,
			y = p->old_y,
			w = p->aux_bmp->w,
			h = p->aux_bmp->h;

		DRS_add_rectangle(x, y, w, h);
		DRS_add_rectangle(x + DISPLAY_W, y, w, h);
		DRS_add_rectangle(x - DISPLAY_W, y, w, h);
		DRS_add_rectangle(x, y + DISPLAY_H, w, h);
		DRS_add_rectangle(x, y - DISPLAY_H, w, h);
	}

	if (p->aux_bmp)
		destroy_bitmap(p->aux_bmp);

	if (p->bmp)
		free(p->bmp);

	remove_particle(p);

	if (p)
		free(p);

	return;
}

void get_particle_velocity(PARTICLE *p, float *vx, float *vy) {
	
	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_velocity: Parameter 'p' is NULL.\n");
		return;
	}
	if (!vx) {
		LOG("--== Error ==-- Particle::get_particle_velocity: Parameter 'vx' is NULL.\n");
		return;
	}
	if (!vy) {
		LOG("--== Error ==-- Particle::get_particle_velocity: Parameter 'vy' is NULL.\n");
		return;
	}

	if (vx)
		*vx = p->vx;
	if (vy)
		*vy = p->vy;	

	return;
}

void get_particle_coord(PARTICLE *p, float *x, float *y) {
	
	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_coord: Parameter 'p' is NULL.\n");
		return;
	}

	if (x)
		*x = p->x;
	if (y)
		*y = p->y;	

	return;
}


BITMAP *get_particle_current_frame(PARTICLE *p) {

	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_current_frame: Parameter 'p' is NULL.\n");
		return NULL;
	}
	if (!p->bmp) {
		return NULL;
	}
	if (!p->num_bitmaps) {
		return NULL;
	}
	if ((p->cur_frame < 0) || (p->cur_frame > p->end_frame)) {
		LOG("--== Error ==-- Particle::get_particle_current_frame: Frame out of range! (%i)\n", p->cur_frame);
		return NULL;
	}

	return (p->aux_bmp ? p->aux_bmp : p->bmp[p->cur_frame >> 16]);
}

BITMAP *get_particle_current_bitmap(PARTICLE *p) {

	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_current_bitmap: Parameter 'p' is NULL.\n");
		return NULL;
	}
	if (!p->bmp) {
		return NULL;
	}
	if (!p->num_bitmaps) {
		return NULL;
	}
	if ((p->cur_frame < 0) || (p->cur_frame > p->end_frame)) {
		LOG("--== Error ==-- Particle::get_particle_current_bitmap: Frame out of range! (%i)\n", p->cur_frame);
		return NULL;
	}

	return p->bmp[p->cur_frame >> 16]; 
}


int get_particle_bitmap_w(PARTICLE *p) {

	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_bitmap_w: Parameter 'p' is NULL.\n");
		return 0;
	}
	if (!p->bmp) {
		LOG("--== Error ==-- Particle::get_particle_bitmap_w: Parameter 'p->bmp' is NULL.\n");
		return 0;
	}
	if (!p->num_bitmaps) {
		LOG("--== Error ==-- Particle::get_particle_bitmap_w: Parameter 'p->num_bitmaps' is NULL.\n");
		return 0;
	}

	return p->bmp[0]->w;
}

int get_particle_bitmap_h(PARTICLE *p) {

	if (!p) {
		LOG("--== Error ==-- Particle::get_particle_bitmap_h: Parameter 'p' is NULL.\n");
		return 0;
	}
	if (!p->bmp) {
		LOG("--== Error ==-- Particle::get_particle_bitmap_h: Parameter 'p->bmp' is NULL.\n");
		return 0;
	}
	if (!p->num_bitmaps) {
		LOG("--== Error ==-- Particle::get_particle_bitmap_h: Parameter 'p->num_bitmaps' is NULL.\n");
		return 0;
	}

	return p->bmp[0]->h;
}



void set_particle_frame(PARTICLE *p, float frame) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_frame: Parameter 'p' is NULL.\n");
		return;
	}

	p->cur_frame = MID(0, ftofix(frame), itofix(p->num_bitmaps) - 1);

	return;
}



void set_particle_frames(PARTICLE *p, float start, float end) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_frames: Parameter 'p' is NULL.\n");
		return;
	}

	p->start_frame = MID(0, ftofix(start), itofix(p->num_bitmaps) - 1);
	p->end_frame   = MID(0, ftofix(end), itofix(p->num_bitmaps) - 1);

	return;
}


void set_particle_anim_speed(PARTICLE *p, float speed) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_anim_speed: Parameter 'p' is NULL.\n");
		return;
	}

	p->anim_speed = ftofix(speed);

	return;
}




void set_particle_osc_v(PARTICLE *p, int mag, float speed) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_osc_v: Parameter 'p' is NULL.\n");
		return;
	}

	p->osc_v_mag = mag;
	p->osc_v_speed = ftofix(speed);

	return;
}



void set_particle_osc_h(PARTICLE *p, int mag, float speed) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_osc_h: Parameter 'p' is NULL.\n");
		return;
	}

	p->osc_h_mag = mag;
	p->osc_h_speed = ftofix(speed);

	return;
}

void set_particle_power(PARTICLE *p, int power) {

	if (!p) {
		LOG("--== Error ==-- Particle::set_particle_power: Parameter 'p' is NULL.\n");
		return;
	}

	p->power = power;

	return;
}


int move_particles(void *dummy) {

	int i;

	if (GamePaused)
		return 0;

	for (i = 0; i < NumParticles; i++) {
		PARTICLE *p = Particle[i];

		if (!p)
			continue;

		p->x += p->vx / 100;
		p->y += p->vy / 100;
		p->angle += p->rot_speed / 100;

		if (p->prop & PARTICLE_WRAP) {
			while (p->x > DISPLAY_W)
				p->x -= DISPLAY_W;
			while (p->y > DISPLAY_H)
				p->y -= DISPLAY_H;
			while (p->x < 0)
				p->x += DISPLAY_W;
			while (p->y < 0)
				p->y += DISPLAY_H;
		}
		if (p->prop & PARTICLE_ALIVE) {
			p->life--;

			if (p->life <= 0) {			
				destroy_particle(p);
				i--;
				continue;
			}
		}
		if (p->prop & PARTICLE_ALPHA_DIMINISH) {
			p->alpha += p->alpha_speed;

			if (p->alpha_speed < 0 && p->alpha < 64 && p->prop & PARTICLE_COLLIDE)
				p->prop &= ~ PARTICLE_COLLIDE;
		}

		p->cur_frame += p->anim_speed / 100;
		if ((p->cur_frame >> 16) >= (p->end_frame >> 16))
			p->cur_frame = p->start_frame;

	}

	return 0;
}



void draw_particles(BITMAP *dest) {

	int i;
	int x, y, n = 0;
	BITMAP *temp;
	BITMAP *bmp;
	/* Draw the game */	
	
	for (i = 0; i < NumParticles; i++) {
		PARTICLE *p = Particle[i];

		if (!p)
			continue;

		bmp = get_particle_current_bitmap(p);

		if (!bmp)
			continue;
		
		n = 0;
		if (!p->aux_bmp) {

			int w, h;
			
			if (p->prop & PARTICLE_ROTATES) {
				w = (((int)(MAX(bmp->w, bmp->h) * 1.415) + 31) >> 5) << 5;
				h = w;
			}
			else {
				w = bmp->w;
				h = bmp->h;
			}

			if (p->prop & PARTICLE_ALPHA)
				p->aux_bmp = create_bitmap(w, h);
			else
				p->aux_bmp = GFX_CreateAccelMaskedBitmap(w, h);

			n = 1;
		}

		temp = p->aux_bmp;

		if (!temp)
			continue;

		if (!(p->prop & PARTICLE_ROTATES) && ((n) || (p->last_frame != (p->cur_frame >> 16)))) {
			clear_to_color(temp, temp->vtable->mask_color);
			draw_sprite(temp, bmp, 0, 0);
			n = 1;
		}
		else if ((p->prop & PARTICLE_ROTATES) && (ABS(p->old_angle - p->angle) > ftofix(0.5) || (n) || (p->last_frame != (p->cur_frame >> 16)))) {
			clear_to_color(temp, temp->vtable->mask_color);
			GFX_HQ_RotateSprite(temp, bmp, (temp->w - bmp->w) / 2, (temp->h - bmp->h) / 2, p->angle + ftofix(M_PI / 2));
			p->old_angle = p->angle;
			p->last_frame = p->cur_frame >> 16;
			n = 1;
		}
		if (p->prop != p->old_prop) {
			p->old_prop = p->prop;
			n = 1;
		}

		x = (int)p->x - temp->w / 2;
		y = (int)p->y - temp->h / 2;

		if (x != p->old_x || y != p->old_y || n) {
			DRS_add_rectangle(x, y, temp->w, temp->h);
			DRS_add_rectangle(p->old_x, p->old_y, temp->w, temp->h);

			if (p->prop & PARTICLE_WRAP) {			
				if (((p->x + temp->w / 2) >= DISPLAY_W) || ((p->x - temp->w / 2) < 0)) {
					DRS_add_rectangle(x - DISPLAY_W, y, temp->w, temp->h);
					DRS_add_rectangle(x + DISPLAY_W, y, temp->w, temp->h);
					DRS_add_rectangle(p->old_x - DISPLAY_W, p->old_y, temp->w, temp->h);
					DRS_add_rectangle(p->old_x + DISPLAY_W, p->old_y, temp->w, temp->h);
				}
				if (((p->y + temp->h / 2) >= DISPLAY_H) || ((p->y - temp->h / 2) < 0)) {
					DRS_add_rectangle(x, y - DISPLAY_H, temp->w, temp->h);
					DRS_add_rectangle(x, y + DISPLAY_H, temp->w, temp->h);
					DRS_add_rectangle(p->old_x, p->old_y - DISPLAY_H, temp->w, temp->h);
					DRS_add_rectangle(p->old_x, p->old_y + DISPLAY_H, temp->w, temp->h);
				}
			}
		}

		if (p->prop & PARTICLE_INVISIBLE) {
		}
		else if (p->prop & PARTICLE_ALPHA) {

			COLOR_MAP *old_map = color_map;
			int alpha = MIN(p->alpha, 255);

			if (p->alpha <= 0)
				continue;

			if (bitmap_color_depth(screen) == 8)
				color_map = &transmaparray[(alpha << 2) >> 8];
			else
				set_add_blender(128, 128, 128, alpha);

			draw_trans_sprite(dest, temp, x, y);

			if (p->prop & PARTICLE_WRAP) {
				if (((p->x + temp->w / 2) >= DISPLAY_W) || ((p->x - temp->w / 2) < 0)) {
					draw_trans_sprite(dest, temp, x - DISPLAY_W, y);
					draw_trans_sprite(dest, temp, x + DISPLAY_W, y);
				}
				if (((p->y + temp->h / 2) >= DISPLAY_H) || ((p->y - temp->h / 2) < 0)) {
					draw_trans_sprite(dest, temp, x, y - DISPLAY_H);
					draw_trans_sprite(dest, temp, x, y + DISPLAY_H);
				}
			}
			color_map = old_map;
		}
		else {		
			masked_blit(temp, dest, 0, 0, x, y, temp->w, temp->h);

			if (p->prop & PARTICLE_WRAP) {
			
				if (((p->x + temp->w / 2) >= DISPLAY_W) || ((p->x - temp->w / 2) < 0)) {
					masked_blit(temp, dest, 0, 0, x - DISPLAY_W, y, temp->w, temp->h);
					masked_blit(temp, dest, 0, 0, x + DISPLAY_W, y, temp->w, temp->h);
				}
				if (((p->y + temp->h / 2) >= DISPLAY_H) || ((p->y - temp->h / 2) < 0)) {
					masked_blit(temp, dest, 0, 0, x, y - DISPLAY_H, temp->w, temp->h);
					masked_blit(temp, dest, 0, 0, x, y + DISPLAY_H, temp->w, temp->h);
				}
			}
		}
		p->old_x = x;
		p->old_y = y;
	}
	return;
}

void clear_particles() {

	while (NumParticles)
		destroy_particle(Particle[0]);

	if (Particle)
		free(Particle);

	Particle = NULL;
	ArraySize = 0;

	return;
}

int install_particle() {

	if (ParticleInstalled)
		return 0;

	GTIME_AddFunction(move_particles, 10, NULL, GTIME_NORMAL);

	ParticleInstalled = 1;

	return 0;
}

