/* player.c - Player management module
 * 
 * Copyright 2001 by Robert J Ohannessian.
 *
 * This code is placed under the GNU General Public License.
 * Please refer to the accompanying file 'copying.txt' for details.
 */
#include <string.h>

#include "allegro.h"
#include "particle.h"
#include "weapon.h"
#include "player.h"
#include "input.h"
#include "config.h"
#include "log.h"
#include "game.h"


#include "gametimer.h"


static PLAYER **Player = NULL;
static int NumPlayers = 0;
static int ArraySize = 0;
static int PlayerInstalled = 0;


static void add_player(PLAYER *p) {

	if ((ArraySize - NumPlayers) <= 1) {
		PLAYER **pl = realloc(Player, sizeof(PLAYER*) * (ArraySize + 16));

		if (!pl) {
			LOG("--== Error ==-- Player::add_player: Ran out of memory while allocating %i bytes.\n", sizeof(PLAYER*) * (ArraySize + 16));
			return;
		}
		Player = pl;
		ArraySize += 16;
	}

	Player[NumPlayers] = p;

	NumPlayers++;

	return;
}

static void remove_player(PLAYER *p) {

	int i;

	for (i = 0; i < NumPlayers; i++) {
		if (Player[i] == p) {		
			Player[i] = Player[NumPlayers - 1];
			NumPlayers--;
			break;
		}
	}

	return;
}


PLAYER *create_player() {

	PLAYER *p;
	
	p = (PLAYER*)malloc(sizeof(PLAYER));

	if (!p) {
		LOG("--== Error ==-- Player::create_player: Ran out of memory while allocating %i bytes.\n", sizeof(PLAYER));
		return NULL;
	}

	memset(p, 0, sizeof(PLAYER));

	p->Particle = create_particle();

	if (!p->Particle) {
		LOG("--== Error ==-- Player::create_player: Couldn't create particle.\n");
		free(p);
		return NULL;
	}
	set_particle_type(p->Particle, 'P');
	set_particle_property(p->Particle, PARTICLE_WRAP | PARTICLE_COLLIDE | PARTICLE_ROTATES);

	p->EngineVoice = -1;

	add_player(p);

	return p;
}



void destroy_player(PLAYER *p) {

	if (!p) {
		LOG("--== Error ==-- Player::destroy_player: Ran out of memory while allocating %i bytes.\n", sizeof(PLAYER));
		return;
	}

	if (p->Weapon) {
		int i;

		for (i = 0; i < p->NumWeapons; i++)
			if (p->Weapon[i])
				free(p->Weapon[i]);

		free(p->Weapon);
	}

	destroy_particle(p->Particle);
	remove_player(p);

	free(p);

	return;
}


void set_player_bitmap(PLAYER *p, BITMAP *bmp) {

	if (!p) {
		LOG("--== Error ==-- Player::set_player_bitmap: Parameter 'p' was NULL.\n");
		return;
	}

	remove_particle_bitmaps(p->Particle);
	add_particle_bitmap(p->Particle, bmp);

	return;
}

void set_player_coord(PLAYER *p, float x, float y) {

	if (!p) {
		LOG("--== Error ==-- Player::set_player_coord: Parameter 'p' was NULL.\n");
		return;
	}

	set_particle_coord(p->Particle, x, y);

	return;
}

void set_player_property(PLAYER *p, int prop) {

	if (!p) {
		LOG("--== Error ==-- Player::set_player_property: Parameter 'p' was NULL.\n");
		return;
	}

	p->Properties |= prop;

	if (prop & PLAYER_INVINCIBILITY)
		p->Invincible = INVINCIBILITY_TIME;

	return;
}

void unset_player_property(PLAYER *p, int prop) {

	if (!p) {
		LOG("--== Error ==-- Player::unset_player_property: Parameter 'p' was NULL.\n");
		return;
	}

	p->Properties &= ~prop;

	if (prop & PLAYER_INVINCIBILITY)
		p->Invincible = 0;

	return;
}


void set_player_invincibility(PLAYER *p, int val) {

	if (!p) {
		LOG("--== Error ==-- Player::set_player_invincibility: Parameter 'p' was NULL.\n");
		return;
	}

	p->Invincible = val;

	return;
}


void get_player_velocity(PLAYER *p, float *vx, float *vy) {
	
	if (!p) {
		LOG("--== Error ==-- Player::get_player_velocity: Parameter 'p' was NULL.\n");
		return;
	}

	get_particle_velocity(p->Particle, vx, vy);

	return;
}


int get_player_property(PLAYER *p, int prop) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_property: Parameter 'p' was NULL.\n");
		return 0;
	}

	return p->Properties & prop;
}


float get_player_angle(PLAYER *p) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_angle: Parameter 'p' was NULL.\n");
		return 0;
	}

	return get_particle_angle(p->Particle);
}


void get_player_coord(PLAYER *p, float *x, float *y) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_coord: Parameter 'p' was NULL.\n");
		return;
	}

	get_particle_coord(p->Particle, x, y);

	return;
}


void set_player_hp(PLAYER *p, int hp, int maxhp) {

	if (!p) {
		LOG("--== Error ==-- Player::set_player_hp: Parameter 'p' was NULL.\n");
		return;
	}

	if (maxhp != -1)
		p->maxhp = maxhp;

	if (hp != -1) {
		p->hp = hp;
		p->hp = MID(0, p->hp, p->maxhp);
	}

	return;
}


int get_player_hp(PLAYER *p) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_hp: Parameter 'p' was NULL.\n");
		return 0;	
	}

	return p->hp;
}

int get_player_maxhp(PLAYER *p) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_maxhp: Parameter 'p' was NULL.\n");
		return 1;
	}

	return (p->maxhp < 1) ? 1 : p->maxhp;
}

void add_player_score(PLAYER *p, int val) {

	if (!p) {
		LOG("--== Error ==-- Player::add_player_score: Parameter 'p' was NULL.\n");
		return;
	}

	p->Score += val;

	return;
}

int get_player_score(PLAYER *p) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_score: Parameter 'p' was NULL.\n");
		return 0;
	}

	return p->Score;
}

PARTICLE *get_player_particle(PLAYER *p) {

	if (!p) {
		LOG("--== Error ==-- Player::get_player_particle: Parameter 'p' was NULL.\n");
		return NULL;
	}

	return p->Particle;
}

void set_player_next_shot_delay(PLAYER *p, int val) {

	if (!p || p->SelectedWeapon >= p->NumWeapons)
		return;

	set_weapon_next_shot_delay(p->Weapon[p->SelectedWeapon], val);

	return;
}

int get_player_next_shot_delay(PLAYER *p) {

	if (!p || p->SelectedWeapon >= p->NumWeapons)
		return 0;

	return get_weapon_next_shot_delay(p->Weapon[p->SelectedWeapon]);
}

int get_player_ammo(PLAYER *p) {

	if (!p || p->SelectedWeapon >= p->NumWeapons)
		return 0;

	return get_ship_weapon_ammo(p->Weapon[p->SelectedWeapon]);
}

int get_player_ammo_percent(PLAYER *p) {
	
	int a;

	if (!p || p->SelectedWeapon >= p->NumWeapons)
		return 0;

	a = get_weapon_ammo(p->Weapon[p->SelectedWeapon]->Weapon);
	a = MAX(1, a);

	return get_ship_weapon_ammo(p->Weapon[p->SelectedWeapon]) * 100 / a;
}



int player_fire(PLAYER *p) {
	if (!p || !p->NumWeapons)
		return FALSE;

	if (get_player_next_shot_delay(p))
		return FALSE;

	weapon_fire(p->Weapon[p->SelectedWeapon], p->Particle);

	/* Disband weapons if ammo = 0 */
	if (p->Weapon[p->SelectedWeapon]->Ammo == 0) {
		remove_player_weapon(p, p->Weapon[p->SelectedWeapon]->Weapon);		
		p->SelectedWeapon = MID(0, p->SelectedWeapon, p->NumWeapons - 1);
	}



	return TRUE;
}

void add_player_weapon(PLAYER *p, WEAPON *w) {

	if (!p || !w)
		return;

	p->Weapon = realloc(p->Weapon, sizeof(SHIP_WEAPON*) * (p->NumWeapons + 1));

	if (!p->Weapon)
		return;

	p->Weapon[p->NumWeapons] = malloc(sizeof(SHIP_WEAPON));

	if (!p->Weapon[p->NumWeapons])
		return;

	p->Weapon[p->NumWeapons]->ShootDelay = 0;
	p->Weapon[p->NumWeapons]->Ammo = w->Ammo;
	p->Weapon[p->NumWeapons]->Weapon = w;

	p->NumWeapons++;

	return;
}


void remove_player_weapon(PLAYER *p, WEAPON *w) {

	int i, j;

	if (!p || !w)
		return;	

	for (i = 0; i < p->NumWeapons; i++) {

		if (p->Weapon[i]->Weapon == w) {

			free(p->Weapon[i]);

			for (j = i; j < p->NumWeapons - 1; j++)
				p->Weapon[j] = p->Weapon[j + 1];

			p->NumWeapons--;

			if (p->NumWeapons) {
				SHIP_WEAPON **s = realloc(p->Weapon, sizeof(SHIP_WEAPON*) * p->NumWeapons);

				if (s)
					p->Weapon = s;
			}
			else {
				free(p->Weapon);
				p->Weapon = NULL;
			}
		}
	}

	return;
}


int player_has_weapon(PLAYER *p, WEAPON *w) {

	int i;

	if (!p || !w)
		FALSE;

	for (i = 0; i < p->NumWeapons; i++) {

		if (p->Weapon[i]->Weapon == w)
			return TRUE;
	}

	return FALSE;
}


void remove_all_player_weapons(PLAYER *p) {

	if (!p)
		return;

	while (p->NumWeapons)
		remove_player_weapon(p, p->Weapon[0]->Weapon);

	return;
}

void select_player_weapon(PLAYER *p, WEAPON *w) {

	int i;

	if (!p)
		return;

	for (i = 0; i < p->NumWeapons; i++) {
		if (p->Weapon[i]->Weapon == w) {
			p->SelectedWeapon = i;
			break;
		}
	}
	return;
}

WEAPON *get_player_weapon(PLAYER *p) {

	if (!p)
		return NULL;

	return p->Weapon[p->SelectedWeapon]->Weapon;
}


void clear_players() {

	while(NumPlayers)
		destroy_player(Player[0]);

	if (Player)
		free(Player);
	Player = NULL;
	ArraySize = 0;

	return;
}


int manage_players(void *dummy) {

	int i;

	if (GamePaused)
		return 0;

	for (i = 0; i < NumPlayers; i++) {

		PLAYER *p = Player[i];

		if (!p)
			continue;

		if (p->Properties & PLAYER_INVINCIBILITY) {
			if (p->Invincible)
				p->Invincible--;
			else
				p->Properties &= ~PLAYER_INVINCIBILITY;
		}

		if (get_player_next_shot_delay(p))
			set_player_next_shot_delay(p, get_player_next_shot_delay(p) - 1);
	}

	return 0;
}

int install_player() {

	if (PlayerInstalled)
		return 0;

	GTIME_AddFunction(manage_players, 10, NULL, GTIME_NORMAL);

	PlayerInstalled = 1;

	return 0;
}


int player_get_input_up(PLAYER *p) {
	return get_input(p->Input, INPUT_KEY_UP);
}
int player_get_input_down(PLAYER *p) {
	return get_input(p->Input, INPUT_KEY_DOWN);
}
int player_get_input_right(PLAYER *p) {
	return get_input(p->Input, INPUT_KEY_RIGHT);
}
int player_get_input_left(PLAYER *p) {
	return get_input(p->Input, INPUT_KEY_LEFT);
}
int player_get_input_fire(PLAYER *p) {
	return get_input(p->Input, INPUT_KEY_FIRE);
}


void load_player_input(PLAYER *p, int n) {
	
	if (!p)
		return;

	if (p->Input)
		destroy_input(&p->Input);

	p->Input = load_player_input_config(n + 1);

	return;
}
